{
  ZPAQ's Face V0.9 BETA Data Compression

    Copyright (C) 2020 Alexander Staidl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3 as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, process, Forms, Controls, Graphics, Dialogs, StdCtrls, math,
  ExtCtrls, ComCtrls, FileCtrl, EditBtn, ShellCtrls, FileUtil, LazFileUtils, StrUtils
  {$IFDEF UNIX},baseunix{$ENDIF UNIX};

type

  { THauptfenster }

  THauptfenster = class(TForm)
    ArchivOeffnenButton: TButton;
    Experteneinstellung: TCheckBox;
    ThreadAnzahl: TComboBox;
    SpeicherBedarf: TComboBox;
    NeuesArchiv: TButton;
    EntpackenButton: TButton;
    Label2: TLabel;
    OrdnerauswahlButton: TButton;
    DateiauswahlButton: TButton;
    Ausgabeaktualisierung: TTimer;
    Label1: TLabel;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Archivinhalt: TShellListView;
    SelectDirectoryDialog1: TSelectDirectoryDialog;
    ZPAQ: TProcess;
    StartButton: TButton;
    Kompressionslevel: TComboBox;
    AusgabeMemo: TMemo;
    procedure ArchivOeffnenButtonClick(Sender: TObject);
    procedure AusgabeaktualisierungTimer(Sender: TObject);
    procedure ExperteneinstellungChange(Sender: TObject);
    procedure DateiauswahlButtonClick(Sender: TObject);
    procedure EntpackenButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure KompressionslevelChange(Sender: TObject);
    procedure NeuesArchivClick(Sender: TObject);
    procedure OrdnerauswahlButtonClick(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    function DateinameOhnePfad(DateinameMitPfad:string): string;                  //Folgende Funktion streift den Dateipfad ab. Kann auch auf Ordner angewendet werden:
    procedure ThreadAnzahlChange(Sender: TObject);
    procedure ZPAQAusgabeEinlesen(TeilweiseEinlesen:boolean);                     //Liest die Ausgabe von ZPAQ in die AusgabeStrLst ein.
    procedure ArchivinhaltAusAusgabeParsen;
    procedure SpeicherbedarfBerechnen(KLevel, ThreadAnz:integer);                 //KLevel=Kompressionslevel
    procedure LetzteAusgabenzeilenInsMemo(Anzahl: integer);
  private

  public

  end;

var
  Hauptfenster: THauptfenster;
  GeoeffnetesArchiv, GeoeffneterOrdner: string;                                   //GeoeffnererOrdner enthält den zuletzt über die Öffnen/hinzufügen-Diagloge genutzten Ordner.
  AusgabeStrLst: TStringList;

implementation

{$R *.lfm}

{ THauptfenster }

procedure THauptfenster.FormCreate(Sender: TObject);
begin
  ZPAQ:=TProcess.Create(nil);
  //ZPAQ.Executable:= FindDefaultExecutablePath('zpaq');                          //Kann verwendet werden, falls zpaq installiert ist.
  ZPAQ.Executable:=ProgramDirectory + 'bin' + DirectorySeparator + 'zpaq'         //Nutzt die im Unterordner "bin" dieser Anwendung befindliche Version von ZPAQ. DirectorySeparator='/' (Linux) bzw. '\' (Windows)
                   {$IFDEF Windows} +'64.exe' {$ENDIF Windows};                   //Unter Windows heißt das Programm zpaq64.exe
  {$IFDEF UNIX} fpChmod(ZPAQ.Executable,&777);  {$ENDIF UNIX}                     //Markiert die zpaq-Binärdatei als ausführbar. Nur unter Unix notwendig.
  ZPAQ.Options := [poUsePipes];                                                   //Für das Auslesen der ZPAQ-Ausgabe
  GeoeffnetesArchiv:='';
  AusgabeStrLst:= TStringList.Create;
  AusgabeStrLst.Clear;
end;

procedure THauptfenster.KompressionslevelChange(Sender: TObject);
begin
  If Experteneinstellung.Checked then
    SpeicherbedarfBerechnen(Kompressionslevel.ItemIndex,ThreadAnzahl.ItemIndex);  //Der Speicherbedarf ist abhängig vom Kompressionslevel und wird hier entsprechend angepasst
end;

procedure THauptfenster.NeuesArchivClick(Sender: TObject);
begin
  GeoeffnetesArchiv:='';                                                          //Zurücksetzen des geöffneten Archivs
  Label1.Caption:='Archivinhalt:';                                                //Zusücksetzen der Beschriftung
  Archivinhalt.Clear;                                                             //Leeren des Archivinhalts
  EntpackenButton.Enabled:=false;                                                 //Buttons werden deaktiviert.
  StartButton.Enabled:=false;
  NeuesArchiv.Enabled:=false;
end;

procedure THauptfenster.OrdnerauswahlButtonClick(Sender: TObject);
var NeuerEintrag: TListItem;
    n: Integer;
begin
  If GeoeffneterOrdner<>'' then SelectDirectoryDialog1.InitialDir:=GeoeffneterOrdner;        //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If SelectDirectoryDialog1.Execute then for n:=0 to SelectDirectoryDialog1.Files.Count-1 do begin  // SelectDirectoryDialog1.Execute öffnet den Ordner-Auswahldialog:
    NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                           //Neuer Eintrag in den Archivinhalt (=TShellListView). Free notwendig oder automatisch mit ShellListView?
    NeuerEintrag.Caption:=SelectDirectoryDialog1.Files.Strings[n];                //Pfad + Ordnername als Caption in die 1. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add('-');                                               //Die Ordnergröße wird nicht angegeben, daher ein '-' in die 2. Spalte des Archivinhalts. ---> ggf. zu späteren Zeitpunkt noch ändern.
    NeuerEintrag.SubItems.Add('Ordner');                                          //Als Typ nun der Eintrag 'Ordner' in die 3. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(DateinameOhnePfad(NeuerEintrag.Caption));           //In die später nicht sichtbare 4. Spalte des Archivinhalts kommt der Ordnername ohne Pfad, damit er später ohne Pfad in das Archiv gespeichert wird.
    Archivinhalt.Items.AddItem(NeuerEintrag);                                     //Jetzt wird der neue Eintrag dem Archivinhalt hinzugefügt:
    GeoeffneterOrdner:=leftstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(NeuerEintrag.SubItems.Strings[2]));  //Streift über leftstr den Ordner des ausgewählten Ordners ab.
    EntpackenButton.Enabled:=false;                                               //Entpacken macht keinen Sinn, wenn Dateien/Ordner hinzugefügt wurden.
    StartButton.Enabled:=true;                                                    //Mit hinzugefügter Datei/Ordner kann das Archiv "erweitert" werden.
    NeuesArchiv.Enabled:=true;
  end;
end;

procedure THauptfenster.StartButtonClick(Sender: TObject);
var n:integer;
begin
  If ZPAQ.Running then begin                                                      //D.h.: "Stopp!" wurde gedrückt. ZPAQ wird also nachfolgend angehalten.
    Ausgabeaktualisierung.Enabled:=false;                                         //Timer wird aus gestellt
    ZPAQ.Terminate(0);                                                            //ZPAQ wird angehalten
    EntpackenButton.Enabled:=false;                                               //EndpackenButton wird deaktiviert, da das Archiv vermutlich wegen der Unterbrechung defekt ist.
    ArchivOeffnenButton.Enabled:=true;                                            //Das Öffnen eine Archivs soll wieder möglich sein.
    StartButton.Caption:='Archiv speichern...';                                   //StartButton wird wieder umbenannt.
    Screen.Cursor:=crDefault;                                                     //Mauszeiger wird zurückgesetzt.
  end else if Archivinhalt.Items.Count>0 then begin                               //Falls das Archiv nicht leer ist, beginnt der Speichervorgang.
    //SAVEDIALOG ANSTEUERN:
    If GeoeffnetesArchiv='' then begin                                            //Globale Variable GeoeffnetesArchiv beinhaltet den Pfad und Dateinamen des angezeigten Archivinhalts.
      If GeoeffnetesArchiv<>''then SaveDialog1.InitialDir:=GeoeffneterOrdner;
      If Archivinhalt.Items.Count=1 then SaveDialog1.FileName:=Archivinhalt.Items.Item[0].SubItems.Strings[2] + '.zpaq' //Falls im Archivinhalt nur 1 Datei/Ordner ist, soll im Save-dialog der Name der Datei/Ordner mit .zpaq-Endung bereits vorgegeben werden.
        else SaveDialog1.FileName:='Unbenanntes_Archiv.zpaq';
      If SaveDialog1.Execute then GeoeffnetesArchiv:=SaveDialog1.FileName;
    end;

    If GeoeffnetesArchiv<>'' then begin
      AusgabeStrLst.Clear;                                                          //Löscht den Inhalt der AusgabeStrLst

      //ZPAQ ANSTEUERUNG:
      ZPAQ.Parameters.Clear;                                                      //Die Parameter, die ev. noch vorhanden sind, werden geleert:
      With ZPAQ.Parameters do begin                                               //Jetzt folgt die Hinzufügung der Programmparameter:
        add('a');                                                                 //Der Parameter a (=add) sagt ZPAQ, es soll die folgenden Inhalte dem Archiv hinzufügen:
        add(GeoeffnetesArchiv);                                                   //Anschließend wird der Name des Archivs an ZPAQ übergeben:
        for n:=0 to Archivinhalt.Items.Count-1 do                                 //Hier wird ZPAQ die Liste der zu komprimierenden Inhalte übergeben:
          add(Archivinhalt.Items.Item[n].Caption);
        add('-to');                                                               //Nun wird ZPAQ angewiesen, diese Dateien/Ordner OHNE Pfad im Verzeichnis abzuspeichern. Dabei befinden sich die Namen ohne Pfad bereits im 3. SubItem des Archivinhalts:
        for n:=0 to Archivinhalt.Items.Count-1 do
          add(Archivinhalt.Items.Item[n].SubItems.Strings[2]);                    //[2], da ab 0 gezählt wird
        If not Experteneinstellung.Checked then add('-m' + IntToStr(Kompressionslevel.ItemIndex)) else begin //Schließlich wird das Kompressionslevel übermittelt:
          add('-m' + IntToStr(Kompressionslevel.ItemIndex) + IntToStr(SpeicherBedarf.ItemIndex));  //Experteneinstellung: BlockSize wird zusätzlich an ZPAQ übermittelt.
          add('-t' + IntToStr(ThreadAnzahl.ItemIndex));                           //Experteneinstellung: Threadanzahl wird zusätzlich übermittelt.
        end; //Experteneinstellung
        add('-s1');
      end; //with Parameters
      ZPAQ.Execute;                                                               //Ausführen von ZPAQ mit den obigen Parametern:

      AusgabeStrLst.Clear;                                                          //Leeren der AusgabeStrLst
      Label1.Caption:='Archivinhalt: ' + GeoeffnetesArchiv;                       //Dateiname wird im Label angezeigt.
      EntpackenButton.Enabled:=false;                                             //Wird erst nach Abschluss des Komprimiervorgangs wieder aktiviert-
      ArchivOeffnenButton.Enabled:=false;                                         // ''
      StartButton.Caption:='STOPP!';                                              //Der Komprimiervorgang soll durch erneuten Klick auf die Schaltfläche angehalten werden können, hier die Umbenennung:
      Screen.Cursor:=crHourGlass;                                                  //Mauszeiger wird mit Uhrenglas versehen.

      Ausgabeaktualisierung.Enabled:=true;                                        //Der Timer Ausgabeaktualisierung aktualisiert das Fenster für die Konsolenausgabe:
    end else showmessage('Leerer Dateiname - das Archiv konnte nicht gespeichert werden.');
  end;
end; //StartButtonClick

procedure THauptfenster.FormClose(Sender: TObject; var CloseAction: TCloseAction
  );
begin
  If ZPAQ.Running then ZPAQ.Terminate(0);                                         //ZPAQ wird abgewürgt, falls es noch läuft.
  ZPAQ.Free;                                                                      //Freigeben des TProcess.
  AusgabeStrLst.Free;
end;

procedure THauptfenster.AusgabeaktualisierungTimer(Sender: TObject);
begin
  ZPAQAusgabeEinlesen(true);                                                      //Einlesen der ZPAQ-Ausgabe in die AusgabeStrLst
  If not ZPAQ.Running then begin                                                  //Falls ZPAQ fertig komprimiert hat...
    Ausgabeaktualisierung.Enabled:=false;                                         //Timer wird deaktiviert
    StartButton.Caption:='Archiv speichern...';                                   //StartButton wird zurückgesetzt
    EntpackenButton.Enabled:=true;                                                //Archiv kann also auch direkt wieder entpackt werden...
    EntpackenButton.Caption:='Archiv entpacken...';                               //EntpackenButton wird zurückgesetzt.
    ArchivOeffnenButton.Enabled:=true;                                            //Das Öffnen eines Archivs ist jetzt wieder möglich.
    Screen.Cursor:=crDefault;                                                     //Mauszeiger wird zurückgesetzt.
    //(Zeile oben muss ggf. noch angepasst werden für den Fall, dass jemand das Komprimieren abbricht und die Datei somit nicht fehlerfrei existiert.
  end;
  LetzteAusgabenzeilenInsMemo(10);
  application.ProcessMessages;                                                  //Tipp von winni. Ergab bei mir keine Änderung, funktioniert allerdings bei Forennutzer winni unter KDE.
end;

procedure THauptfenster.ExperteneinstellungChange(Sender: TObject);
begin
  If Experteneinstellung.Checked then begin
    ThreadAnzahl.Enabled := true;
    SpeicherBedarf.Enabled := true;
    SpeicherbedarfBerechnen(Kompressionslevel.ItemIndex,ThreadAnzahl.ItemIndex);
  end else
  begin
    ThreadAnzahl.Enabled := false;
    ThreadAnzahl.ItemIndex:=0;
    SpeicherBedarf.Enabled := false;
    SpeicherBedarf.Clear;
  end;
end;

procedure THauptfenster.ArchivOeffnenButtonClick(Sender: TObject);
var n:integer;
begin
  If GeoeffneterOrdner<>'' then OpenDialog1.InitialDir:=GeoeffneterOrdner;        //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If not ZPAQ.Running then if OpenDialog1.Execute then
    if length(OpenDialog1.FileName)>5 then if rightstr(OpenDialog1.FileName,5)='.zpaq' then begin
    AusgabeStrLst.Clear;
    Archivinhalt.Clear;                                                           //Archivinhalt wird gelöscht
    With ZPAQ.Parameters do begin
      Clear;                                                                      //Zunächst werden die Parameter bereinigt.
      add('l');                                                                   //Mit der Anweisung 'l' listet ZPAQ den Archivinhalt auf.
      add(OpenDialog1.FileName);                                                  //Übergabe des Dateinamens
    end;
    ZPAQ.Execute;                                                                 //ZPAQ beginnt mit der Auflistung des Inhalts.

    ZPAQAusgabeEinlesen(false);                                                   //false bedeutet, dass die Ausgabe von ZPAQ vollständig, d.h. bis ZPAQ beendet ist, eingelesen wird.

    ArchivinhaltAusAusgabeParsen;                                                 //Parsen der AusgabeStrLst. Geparster Inhalt wird in dem Archivinhalt gespreichert.

    LetzteAusgabenzeilenInsMemo(100);

    GeoeffnetesArchiv:=OpenDialog1.FileName;                                      //Globale Variable beinhaltet Pfad + Namen des geöffneten Archivs.
    GeoeffneterOrdner:=leftstr(GeoeffnetesArchiv,length(GeoeffnetesArchiv)-length(DateinameOhnePfad(GeoeffnetesArchiv)));
    Label1.Caption:='Archivinhalt: ' + GeoeffnetesArchiv;                         //Label1 zeigt den Archivnamen an.
    EntpackenButton.Enabled:=true;                                                //Archiv kann also auch entpackt werden.
    StartButton.Enabled:=false;                                                   //Direkt geöffnetes Archiv muss nicht direkt wieder gespeichert werden.
    NeuesArchiv.Enabled:=true;                                                    //Buttons werden aktiviert/deaktiviert

  end else showmessage('Ausgewählte Datei konnte nicht geöffnet werden. Die Endung muss .zpaq sein.');
end;

procedure THauptfenster.DateiauswahlButtonClick(Sender: TObject);
var NeuerEintrag: TListItem;
    n: Integer;
begin
  If GeoeffneterOrdner<>'' then OpenDialog1.InitialDir:=GeoeffneterOrdner;        //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If OpenDialog1.Execute then for n:=0 to OpenDialog1.Files.Count-1 do begin
    NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                           //Neuer Eintrag in den Archivinhalt (=TShellListView). Frage: Free notwendig oder automatisch mit ShellListView?
    NeuerEintrag.Caption:=OpenDialog1.Files.Strings[n];                           //Pfad + Dateiname als Caption in die 1. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(FloatToStr(FileSize(NeuerEintrag.Caption)/1000) + 'kB');  //Dateigröße in kB als SubItem für die 2. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(rightstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(ExtractFileNameWithoutExt(NeuerEintrag.Caption))));  //Dateiendung als SubItem für die 3. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(DateinameOhnePfad(NeuerEintrag.Caption));           //In die später nicht sichtbare 4. Spalte des Archivinhalts kommt der Dateiname ohne Pfad, damit sie später ohne Pfad in das Archiv gespeichert wird.
    Archivinhalt.Items.AddItem(NeuerEintrag);                                     //Jetzt wird der neue Eintrag dem Archivinhalt hinzugefügt:
    GeoeffneterOrdner:=leftstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(NeuerEintrag.SubItems.Strings[2]));  //Streift über leftstr den Ordner der neu geöffneten Datei ab.
    EntpackenButton.Enabled:=false;                                               //Entpacken macht keinen Sinn, wenn Dateien/Ordner hinzugefügt wurden.
    StartButton.Enabled:=true;                                                    //Mit hinzugefügter Datei/Ordner kann das Archiv "erweitert" werden.
    NeuesArchiv.Enabled:=true;
  end;
end;

procedure THauptfenster.EntpackenButtonClick(Sender: TObject);
begin
  If GeoeffneterOrdner<>'' then SelectDirectoryDialog1.InitialDir:=GeoeffneterOrdner;   //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If (GeoeffnetesArchiv<>'') and (not ZPAQ.Running) then                          //Das Entpacken geht nur, wenn ZPAQ nicht schon läuft und wenn das Archiv geladen ist.
    if SelectDirectoryDialog1.Execute then begin                                  //Ordner zum Entpacken wird gewählt
    EntpackenButton.Caption:='ENTPACKEN LÄUFT...';                                //Damit Nutzer sieht, dass das Programm arbeitet
    EntpackenButton.Enabled:=false;                                               //Während des Entpackens sollen die ZPAQ-nutzenden Buttons deaktiviert werden-
    ArchivOeffnenButton.Enabled:=false;
    GeoeffneterOrdner:=SelectDirectoryDialog1.FileName + DirectorySeparator;      //Die Variable für den zuletzt geöffneten Ordner wird auf das Zielverzeichnis gesetzt. DirectoryDeparator hängt noch ein "/" an.

    //ZPAQ Ansteuerung:
    With ZPAQ.Parameters do begin
      clear;                                                                      //ZPAQ-Parameter zurücksetzen
      add('x');                                                                   //Komando x lässt das Archiv extrahieren
      add(GeoeffnetesArchiv);                                                     //Übergabe des Pfads und der Datei des zu entkomprimierenden Archivs
      add('-to');                                                                 //Befehl, wohin das Archiv gespeichert werden soll
      add(SelectDirectoryDialog1.FileName);                                       //Filename enthält den Pfad des Ordners, in den das Archiv entpackt werden soll.
    end;
    ZPAQ.Execute;                                                                 //Entpackvorgang wird gestartet

    Ausgabeaktualisierung.Enabled:=true;                                          //Der Output von ZPAQ wird in die Konsole gespeichert.
  end;
end;

//EIGENE FUNKTIONEN UND PROZEDUREN:

//Folgende Funktion streift den Pfad von Dateien UND Ordnern ab.
function THauptfenster.DateinameOhnePfad(DateinameMitPfad:string): string;
var n: integer;
begin
  RESULT:=DateinameMitPfad;                                                       //Falls die Datei bereits ohne Pfad vorliegt, wird der Rückgabewert im Vorfeld initialisiert als:
  n:=length(DateinameMitPfad)-1;                                                  //Der DateinameMitPfad wird von hinten nach vorne untersucht. Das hinterste Zeichen wird nicht berücksichtigt, damit die Kompatibilität zu Ordnern erhalten bleibt.
  while (n>0) and (DateiNameMitPfad[n]<>DirectorySeparator) do dec(n);            //Das Ende des vorangestellten Pfades wird durch das Symbol '/' erkannt (bzw. unter Windows mit '\'=DirectorySeparator)
  If DateinameMitPfad[n]=DirectorySeparator then                                  //Hier wird nun der Pfad abgestreift, falls er vorhanden ist:
    DateinameOhnePfad:=rightstr(DateinameMitPfad, length(DateinameMitPfad)-n);
end; //DateinameOhnePfad

procedure THauptfenster.ThreadAnzahlChange(Sender: TObject);
begin
  SpeicherbedarfBerechnen(Kompressionslevel.ItemIndex,ThreadAnzahl.ItemIndex);
end;

procedure THauptfenster.ZPAQAusgabeEinlesen(TeilweiseEinlesen:boolean);
var
  OutputStream : TStream;                                                         //In diesen Stream wird die Ausgabe von ZPAQ gespeichert.
  NeueZeilen: TStringList;
  BytesRead, VerfuegbareBytes: longint;                                           //Variablen, die speichern, wie groß der Output ist.
  Buffer       : array of byte;                                                   //Hier wird die Ausgabe von ZPAQ zwischengespeichert.

begin
  OutputStream := TMemoryStream.Create;
  NeueZeilen := TStringList.Create;                                               //NeueZeilen ist die Liste mit den neuen Zeilen der ZPAQ-Ausgabe
  repeat
    VerfuegbareBytes:=ZPAQ.Output.NumBytesAvailable;                              //Speichert, wie viele verfügbare Bytes ZPAQ ausgibt.
    setlength(Buffer, VerfuegbareBytes);                                          //Die Puffergröße wird entsprechend gesetzt
    BytesRead := ZPAQ.Output.Read(Buffer[0], VerfuegbareBytes);                   //Die ZPAQ-Ausgabe wird in den Buffer übertragen...
    OutputStream.Write(Buffer[0], BytesRead);                                     //...und vom Buffer in den OutPutStream übertragen.
  until (BytesRead = 0)and((ZPAQ.Running=false) or (TeilweiseEinlesen=true));     //Falls TeilweiseEinlesen=false, dann läuft die Schleife so lange, bis ZPAQ beendet ist.
  OutputStream.Position:=0;                                                       //Der OutputStream soll anschließend beginnend von Position 0 übergeben werden.
  NeueZeilen.LoadFromStream(OutputStream);                                        //Der OutputStream wird in NeueZeilen gespeichert
  AusgabeStrLst.AddStrings(NeueZeilen);                                       //Die neuen Zeilen werden der AusgabeStrLst hinzugefügt

  OutputStream.Free;                                                              //Freigabe des OutputStream.
  NeueZeilen.Free;                                                                //Freigabe von NeueZeilen
end;

procedure THauptfenster.ArchivinhaltAusAusgabeParsen;
var Zeile, n: integer;
    ZwSpStr: string;
    NeuerEintrag: TListItem;
begin
  Zeile:=0;
  ZwSpStr:='';
  while Zeile<AusgabeStrLst.Count do begin
    If length(AusgabeStrLst.Strings[Zeile])>0 then                            //Manche Zeilen sind Leerzeilen, die nicht geparst werden sollen.
      if (AusgabeStrLst.Strings[Zeile][1]='-')and(length(AusgabeStrLst.Strings[Zeile])>42) then begin   //ZPAQ-Ausgabezeilen mit Dateiauflistung beginnen mit einem "-". Um Einlesefehler abzufangen wird noch geprüft, ob mehr als 42 Zeichen in der Zeile sind - ansonsten gibt es unten Zugriffsfehler.
        NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                       //Für einen neuen Eintrag im Archivinhalt
        NeuerEintrag.SubItems.Add('');                                            //Für Dateigröße
        NeuerEintrag.SubItems.Add('');                                            //Für Dateityp
        NeuerEintrag.SubItems.Add('');                                            //Für Datei ohne Pfad
                                                                                  //Aufbau:       Datum    Uhrzeit   Dateigröße d=Ordner Pfad/DateiOderOrdner.MöglicheEndung
                                                                                  //Aufbau: "- 2020-07-16 16:27:46       747977 d0775 lib/x86_64-linux/" im Fall eines Ordners: d=Directory
                                                                                  //Aufbau: "- 2020-07-15 10:52:44         2048  0664 /home/alexander/Schreibtisch/ZPAQ-FACE/unit1.lfm" im Fall einer Datei
        ZwSpStr:=midstr(AusgabeStrLst.Strings[Zeile],42, length(AusgabeStrLst.Strings[Zeile]));  //Ab ab dem 42. Zeichen beginnt immer der Dateiname+Pfad
        NeuerEintrag.Caption:=ZwSpStr;                                            //Pfad + Dateiname werden an den neuen Eintrag übergeben.
        NeuerEintrag.SubItems.Strings[2]:=DateinameOhnePfad(ZwSpStr);             //Nur der Vollständigkeit halber: In nicht sichtbarer 4. Spalte der Tabelle die Datei/der Ordner OHNE Pfad.
        ZwSpStr:='';

        If AusgabeStrLst.Strings[Zeile][36]='d' then NeuerEintrag.SubItems.Strings[1]:='Ordner'    //Im Fall eines Ordners steht an 36. Stelle ein d. Bei Dateityp wird "Ordner" angezeigt, falls es sich um einen Ordner handelt, ansonsten die Dateiendung...
          else NeuerEintrag.SubItems.Strings[1]:=rightstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(ExtractFileNameWithoutExt(NeuerEintrag.Caption)));  //Dateiendung falls vorhanden.

        n:=34;                                                                    //Nachfolgend wird von der 34. Stelle nach hinten bis zum ersten Leerzeichen geparst. Hier steht die Dateigröße.
        ZwSpStr:='';
        repeat
          ZwSpStr:=AusgabeStrLst.Strings[Zeile][n] + ZwSpStr;
          dec(n)
        until (AusgabeStrLst.Strings[Zeile][n]=' ') or (n=1);                 //Es wurde erneut bis zum Leerzeichen geparst. D.h. ZwSpStr enthält jetzt die Dateigröße in Bytes.

        NeuerEintrag.SubItems.Strings[0]:=FloatToStr(StrToInt(ZwSpStr)/1000) + ' kB';  //Dateigröße in kB an den neuen Eintrag übergeben.
        Archivinhalt.Items.AddItem(NeuerEintrag);                                 //Neuer Eintrag wird an den Archivinhalt übergeben.

      end; //if AusgabeStrLst.Strings[Zeile][1]='-'
    inc(Zeile);
  end; //while Zeile<AusgabeStrLst.Count
end; //procedure ArchivinhaltAusAusgabeParsen

procedure THauptfenster.SpeicherbedarfBerechnen(KLevel, ThreadAnz:integer);
var BlockSize:integer;
    Speicher: real;
begin
  SpeicherBedarf.Items.Clear;
  If ThreadAnz = 0 then begin
    for BlockSize:=0 to 11 do
      SpeicherBedarf.Items.Add('BlockSize: ' + IntToStr(BlockSize));
  end else begin
    for BlockSize:=0 to 11 do begin
      If KLevel = 5 then Speicher := 16 * ThreadAnz * power(2,BlockSize) else
      Speicher := 8 * ThreadAnz * power(2,BlockSize);
      SpeicherBedarf.Items.Add(FloatToStr(Speicher) + ' MiB');
    end;
  end;
  SpeicherBedarf.ItemIndex:=0;
end; //SpeicherbedarfBerechnen

procedure THauptfenster.LetzteAusgabenzeilenInsMemo(Anzahl: integer);
var Anfang, n: integer;
begin
  AusgabeMemo.Clear;
  If AusgabeStrLst.Count<Anzahl then Anfang:=0 else Anfang := AusgabeStrLst.Count-Anzahl;
  for n := Anfang to AusgabeStrLst.Count-1 do AusgabeMemo.Lines.Add(AusgabeStrLst.Strings[n]);
  AusgabeMemo.SelStart := Length(AusgabeMemo.Text);                               //Das Ausgabememo wird so eingestellt, dass die letzte Zeile gesehen werden kann.
  AusgabeMemo.Refresh;                                                            //Neuen Inhalt anzeigen.

end;  //LetzteAusgabenzeilenInsMemo

end.

